@@ -1,7 +1,7 @@
 #ifndef SCM_DEPRECATED_H
 #define SCM_DEPRECATED_H
 
-/* Copyright 2003-2007,2009-2018
+/* Copyright 2003-2007,2009-2018,2020
      Free Software Foundation, Inc.
 
    This file is part of Guile.
@@ -111,8 +111,35 @@ typedef ptrdiff_t scm_t_ptrdiff SCM_DEPRECATED_TYPE;
 typedef struct scm_thread scm_i_thread SCM_DEPRECATED_TYPE;
 #undef SCM_DEPRECATED_TYPE
 
+#define SCM_MEMORY_ERROR do { scm_report_out_of_memory (); } while (0)
+
 SCM_DEPRECATED char* scm_find_executable (const char *name);
 
+SCM_DEPRECATED SCM scm_bitvector_p (SCM vec);
+SCM_DEPRECATED SCM scm_bitvector (SCM bits);
+SCM_DEPRECATED SCM scm_make_bitvector (SCM len, SCM fill);
+SCM_DEPRECATED SCM scm_bitvector_length (SCM vec);
+SCM_DEPRECATED SCM scm_c_bitvector_ref (SCM vec, size_t idx);
+SCM_DEPRECATED SCM scm_bitvector_ref (SCM vec, SCM idx);
+SCM_DEPRECATED void scm_c_bitvector_set_x (SCM vec, size_t idx, SCM val);
+SCM_DEPRECATED SCM scm_bitvector_set_x (SCM vec, SCM idx, SCM val);
+SCM_DEPRECATED SCM scm_bitvector_fill_x (SCM vec, SCM val);
+SCM_DEPRECATED SCM scm_bit_invert_x (SCM vec);
+SCM_DEPRECATED SCM scm_bit_count (SCM item, SCM seq);
+SCM_DEPRECATED SCM scm_bit_count_star (SCM v, SCM kv, SCM obj);
+SCM_DEPRECATED SCM scm_bit_position (SCM item, SCM v, SCM k);
+SCM_DEPRECATED SCM scm_bit_set_star_x (SCM v, SCM kv, SCM obj);
+SCM_DEPRECATED SCM scm_istr2bve (SCM str);
+
+#define SCM_SOURCE_PROPERTY_FLAG_BREAK 1
+
+SCM_DEPRECATED scm_t_bits scm_tc16_srcprops;
+SCM_DEPRECATED SCM scm_sym_copy;
+SCM_DEPRECATED SCM scm_make_srcprops (long line, int col, SCM filename,
+                                      SCM copy, SCM alist);
+
+SCM_DEPRECATED SCM scm_copy_tree (SCM obj);
+
 void scm_i_init_deprecated (void);
 
 #endif
