@@ -1,20 +1,21 @@
-/* Copyright (C) 1995,1996,1997,1998,2000,2001,2003, 2004, 2006, 2009, 2010, 2011 Free Software Foundation, Inc.
- * 
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
- * the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
- * 02110-1301 USA
- */
+/* Copyright 1995-1998,2000-2001,2003-2004,2006,2009-2011,2017-2018
+     Free Software Foundation, Inc.
+
+   This file is part of Guile.
+
+   Guile is free software: you can redistribute it and/or modify it
+   under the terms of the GNU Lesser General Public License as published
+   by the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
+
+   Guile is distributed in the hope that it will be useful, but WITHOUT
+   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+   License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with Guile.  If not, see
+   <https://www.gnu.org/licenses/>.  */
 
 
 #ifdef HAVE_CONFIG_H
@@ -22,32 +23,31 @@
 #endif
 
 #include <math.h>
-
-#include "libguile/_scm.h"
-#include "libguile/array-map.h"
-#include "libguile/stackchk.h"
-#include "libguile/strorder.h"
-#include "libguile/async.h"
-#include "libguile/smob.h"
-#include "libguile/arrays.h"
-#include "libguile/vectors.h"
-#include "libguile/hashtab.h"
-#include "libguile/bytevectors.h"
-#include "libguile/syntax.h"
-
-#include "libguile/struct.h"
-#include "libguile/goops.h"
-
-#include "libguile/validate.h"
-#include "libguile/eq.h"
-
-#include "libguile/private-options.h"
-
-
-
-#ifdef HAVE_STRING_H
 #include <string.h>
-#endif
+
+#include "array-map.h"
+#include "async.h"
+#include "bitvectors.h"
+#include "boolean.h"
+#include "bytevectors.h"
+#include "eval.h"
+#include "foreign.h"
+#include "generalized-arrays.h"
+#include "goops.h"
+#include "gsubr.h"
+#include "hashtab.h"
+#include "pairs.h"
+#include "private-options.h"
+#include "smob.h"
+#include "stackchk.h"
+#include "strorder.h"
+#include "struct.h"
+#include "syntax.h"
+#include "vectors.h"
+
+#include "eq.h"
+
+
 
 
 static SCM scm_i_eq_p (SCM x, SCM y, SCM rest);
@@ -156,6 +156,25 @@ scm_i_fraction_equalp (SCM x, SCM y)
 				  SCM_FRACTION_DENOMINATOR (y))));
 }
 
+int
+scm_i_heap_numbers_equal_p (SCM x, SCM y)
+{
+  if (SCM_IMP (x)) abort();
+  switch (SCM_TYP16 (x))
+    {
+    case scm_tc16_big:
+      return scm_is_true (scm_bigequal (x, y));
+    case scm_tc16_real:
+      return scm_is_true (scm_real_equalp (x, y));
+    case scm_tc16_complex:
+      return scm_is_true (scm_complex_equalp (x, y));
+    case scm_tc16_fraction:
+      return scm_is_true (scm_i_fraction_equalp (x, y));
+    default:
+      abort ();
+    }
+}
+
 static SCM scm_i_eqv_p (SCM x, SCM y, SCM rest);
 #include <stdio.h>
 SCM_DEFINE (scm_i_eqv_p, "eqv?", 0, 2, 1,
@@ -210,17 +229,7 @@ SCM scm_eqv_p (SCM x, SCM y)
     default:
       break;
     case scm_tc7_number:
-      switch SCM_TYP16 (x)
-        {
-        case scm_tc16_big:
-          return scm_bigequal (x, y);
-        case scm_tc16_real:
-          return scm_real_equalp (x, y);
-        case scm_tc16_complex:
-          return scm_complex_equalp (x, y);
-	case scm_tc16_fraction:
-          return scm_i_fraction_equalp (x, y);
-        }
+      return scm_from_bool (scm_i_heap_numbers_equal_p (x, y));
     }
   return SCM_BOOL_F;
 }
@@ -396,12 +405,6 @@ scm_equal_p (SCM x, SCM y)
 void
 scm_init_eq ()
 {
-#include "libguile/eq.x"
+#include "eq.x"
 }
 
-
-/*
-  Local Variables:
-  c-file-style: "gnu"
-  End:
-*/
