@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2019  Free Software Foundation, Inc.
+ * Copyright (C) 2012-2020  Free Software Foundation, Inc.
  *
  * This file is part of GNU lightning.
  *
@@ -53,7 +53,7 @@ union jit_pc
 struct jit_literal_pool_entry
 {
   jit_reloc_t reloc;
-  int64_t value;
+  uintptr_t value;
 };
 
 struct jit_literal_pool
@@ -74,6 +74,7 @@ struct jit_state
   uint8_t temp_gpr_saved;
   uint8_t temp_fpr_saved;
   uint8_t overflow;
+  uint8_t emitting_data;
   int frame_size; // Used to know when to align stack.
 #ifdef JIT_NEEDS_LITERAL_POOL
   struct jit_literal_pool *pool;
@@ -96,24 +97,22 @@ static struct jit_literal_pool* alloc_literal_pool(jit_state_t *_jit,
                                                    size_t capacity);
 static void reset_literal_pool(jit_state_t *_jit,
                                struct jit_literal_pool *pool);
-static void grow_literal_pool(jit_state_t *_jit);
-static jit_bool_t add_literal_pool_entry(jit_state_t *_jit,
-                                         struct jit_literal_pool_entry entry,
-                                         uint32_t max_offset);
 static jit_bool_t add_pending_literal(jit_state_t *_jit, jit_reloc_t src,
                                       uint8_t max_offset_bits);
 static void remove_pending_literal(jit_state_t *_jit, jit_reloc_t src);
 static void patch_pending_literal(jit_state_t *_jit, jit_reloc_t src,
-                                  uint64_t value);
+                                  uintptr_t value);
 enum guard_pool { GUARD_NEEDED, NO_GUARD_NEEDED };
 static void emit_literal_pool(jit_state_t *_jit, enum guard_pool guard);
 
 static int32_t read_jmp_offset(uint32_t *loc);
-static int offset_in_jmp_range(ptrdiff_t offset);
+static int offset_in_jmp_range(ptrdiff_t offset, int flags);
 static void patch_jmp_offset(uint32_t *loc, ptrdiff_t offset);
+static void patch_veneer_jmp_offset(uint32_t *loc, ptrdiff_t offset);
 static int32_t read_jcc_offset(uint32_t *loc);
-static int offset_in_jcc_range(ptrdiff_t offset);
+static int offset_in_jcc_range(ptrdiff_t offset, int flags);
 static void patch_jcc_offset(uint32_t *loc, ptrdiff_t offset);
+static void patch_veneer_jcc_offset(uint32_t *loc, ptrdiff_t offset);
 static void patch_veneer(uint32_t *loc, jit_pointer_t addr);
 static int32_t read_load_from_pool_offset(uint32_t *loc);
 #endif
@@ -173,7 +172,8 @@ jit_pointer_t
 jit_address(jit_state_t *_jit)
 {
   ASSERT (_jit->start);
-  return _jit->pc.uc;
+  jit_pointer_t ret = _jit->pc.uc;
+  return _jit->emitting_data ? ret : jit_address_to_function_pointer (ret);
 }
 
 void
@@ -185,6 +185,7 @@ jit_begin(jit_state_t *_jit, uint8_t* buf, size_t length)
   _jit->limit = buf + length;
   _jit->overflow = 0;
   _jit->frame_size = 0;
+  _jit->emitting_data = 0;
 #if JIT_NEEDS_LITERAL_POOL
   ASSERT(_jit->pool->size == 0);
   _jit->pool->deadline = length;
@@ -205,6 +206,7 @@ jit_reset(jit_state_t *_jit)
   _jit->pc.uc = _jit->start = _jit->limit = NULL;
   _jit->overflow = 0;
   _jit->frame_size = 0;
+  _jit->emitting_data = 0;
 #ifdef JIT_NEEDS_LITERAL_POOL
   reset_literal_pool(_jit, _jit->pool);
 #endif
@@ -230,9 +232,10 @@ jit_end(jit_state_t *_jit, size_t *length)
   uint8_t *start = _jit->start;
   uint8_t *end = _jit->pc.uc;
 
-  ASSERT (start);
-  ASSERT (start <= end);
-  ASSERT (end <= _jit->limit);
+  ASSERT(start);
+  ASSERT(start <= end);
+  ASSERT(end <= _jit->limit);
+  ASSERT(!_jit->emitting_data);
 
   jit_flush (start, end);
 
@@ -344,10 +347,17 @@ static inline void emit_u64(jit_state_t *_jit, uint64_t u64) {
   }
 }
 
+static inline void emit_uintptr(jit_state_t *_jit, uintptr_t u) {
+  if (sizeof(u) == 4)
+    emit_u32 (_jit, u);
+  else
+    emit_u64 (_jit, u);
+}
+
 static inline jit_reloc_t
-jit_reloc (jit_state_t *_jit, enum jit_reloc_kind kind,
-           uint8_t inst_start_offset, uint8_t *loc, uint8_t *pc_base,
-           uint8_t rsh)
+jit_reloc(jit_state_t *_jit, enum jit_reloc_kind kind,
+          uint8_t inst_start_offset, uint8_t *loc, uint8_t *pc_base,
+          uint8_t rsh)
 {
   jit_reloc_t ret;
 
@@ -364,6 +374,14 @@ jit_reloc (jit_state_t *_jit, enum jit_reloc_kind kind,
   return ret;
 }
 
+static inline jit_reloc_t
+emit_abs_reloc (jit_state_t *_jit, uint8_t inst_start)
+{
+  uint8_t *loc = _jit->pc.uc;
+  emit_uintptr (_jit, 0);
+  return jit_reloc(_jit, JIT_RELOC_ABSOLUTE, inst_start, loc, _jit->pc.uc, 0);
+}
+
 void
 jit_patch_here(jit_state_t *_jit, jit_reloc_t reloc)
 {
@@ -382,8 +400,11 @@ jit_patch_there(jit_state_t* _jit, jit_reloc_t reloc, jit_pointer_t addr)
   ptrdiff_t diff = (uint8_t*)addr - pc_base;
   ASSERT((diff & ((1 << reloc.rsh) - 1)) == 0);
   diff >>= reloc.rsh;
+#ifdef JIT_NEEDS_LITERAL_POOL
+  int flags = reloc.kind & ~JIT_RELOC_MASK;
+#endif
 
-  switch (reloc.kind)
+  switch (reloc.kind & JIT_RELOC_MASK)
     {
     case JIT_RELOC_ABSOLUTE:
       if (sizeof(diff) == 4)
@@ -408,7 +429,7 @@ jit_patch_there(jit_state_t* _jit, jit_reloc_t reloc, jit_pointer_t addr)
       uint8_t *target = pc_base + (voff << reloc.rsh);
       if (target == loc.uc) {
         // PC still in range to reify direct branch.
-        if (offset_in_jmp_range(diff)) {
+        if (offset_in_jmp_range(diff, flags)) {
           // Target also in range: reify direct branch.
           patch_jmp_offset(loc.ui, diff);
           remove_pending_literal(_jit, reloc);
@@ -427,7 +448,7 @@ jit_patch_there(jit_state_t* _jit, jit_reloc_t reloc, jit_pointer_t addr)
       int32_t voff = read_jcc_offset(loc.ui);
       uint8_t *target = pc_base + (voff << reloc.rsh);
       if (target == loc.uc) {
-        if (offset_in_jcc_range(diff)) {
+        if (offset_in_jcc_range(diff, flags)) {
           patch_jcc_offset(loc.ui, diff);
           remove_pending_literal(_jit, reloc);
         } else {
@@ -466,6 +487,66 @@ jit_patch_there(jit_state_t* _jit, jit_reloc_t reloc, jit_pointer_t addr)
     jit_try_shorten (_jit, reloc, addr);
 }
 
+void
+jit_begin_data(jit_state_t *j, size_t max_size_or_zero)
+{
+#ifdef JIT_NEEDS_LITERAL_POOL
+  if (j->pool->size) {
+    uint8_t *deadline = j->start + j->pool->deadline;
+    // Emit a literal pool now if the data might overwrite the deadline.
+    // Emitting data won't add entries to the pool.
+    if (max_size_or_zero == 0 || j->pc.uc + max_size_or_zero >= deadline)
+      emit_literal_pool(j, NO_GUARD_NEEDED);
+  }
+#endif
+
+  ASSERT(!j->emitting_data);
+  j->emitting_data = 1;
+}
+
+void
+jit_end_data(jit_state_t *j)
+{
+  ASSERT(j->emitting_data);
+  j->emitting_data = 0;
+}
+
+void
+jit_emit_u8(jit_state_t *j, uint8_t u8)
+{
+  ASSERT(j->emitting_data);
+  emit_u8(j, u8);
+}
+
+void
+jit_emit_u16(jit_state_t *j, uint16_t u16)
+{
+  ASSERT(j->emitting_data);
+  emit_u16(j, u16);
+}
+
+void
+jit_emit_u32(jit_state_t *j, uint32_t u32)
+{
+  ASSERT(j->emitting_data);
+  emit_u32(j, u32);
+}
+
+void
+jit_emit_u64(jit_state_t *j, uint64_t u64)
+{
+  ASSERT(j->emitting_data);
+  emit_u64(j, u64);
+}
+
+jit_reloc_t
+jit_emit_addr(jit_state_t *j)
+{
+  ASSERT(j->emitting_data);
+  uint8_t inst_start = 0;
+  return emit_abs_reloc(j, inst_start);
+}
+
 #if defined(__i386__) || defined(__x86_64__)
 # include "x86.c"
 #elif defined(__mips__)
@@ -640,21 +721,26 @@ static void
 abi_gpr_to_mem(jit_state_t *_jit, enum jit_operand_abi abi,
                jit_gpr_t base, ptrdiff_t offset, jit_gpr_t src)
 {
+  // Invariant: GPR memory destination operand sizes are rounded up to words.
+  // True for ARM, AArch64, IA32, and X86-64.  Some ABIs expect to be able to
+  // load operands from the stack via a full-word read, so we need to make sure
+  // we don't leave garbage in the high bytes of (for example) the stack slot
+  // for a uint8_t arg.
   switch (abi) {
   case JIT_OPERAND_ABI_UINT8:
   case JIT_OPERAND_ABI_INT8:
-    jit_stxi_c(_jit, offset, base, src);
+    jit_stxi(_jit, offset, base, src);
     break;
   case JIT_OPERAND_ABI_UINT16:
   case JIT_OPERAND_ABI_INT16:
-    jit_stxi_s(_jit, offset, base, src);
+    jit_stxi(_jit, offset, base, src);
     break;
   case JIT_OPERAND_ABI_UINT32:
   case JIT_OPERAND_ABI_INT32:
 #if __WORDSIZE == 32
   case JIT_OPERAND_ABI_POINTER:
 #endif
-    jit_stxi_i(_jit, offset, base, src);
+    jit_stxi(_jit, offset, base, src);
     break;
 #if __WORDSIZE == 64
   case JIT_OPERAND_ABI_UINT64:
@@ -1242,8 +1328,8 @@ static void
 reset_literal_pool(jit_state_t *_jit, struct jit_literal_pool *pool)
 {
   pool->deadline = _jit->limit - _jit->start;
-  pool->size = 0;
   memset(pool->entries, 0, sizeof(pool->entries[0]) * pool->size);
+  pool->size = 0;
 }
 
 #define INITIAL_LITERAL_POOL_CAPACITY 12
@@ -1279,9 +1365,11 @@ static jit_bool_t
 add_literal_pool_entry(jit_state_t *_jit, struct jit_literal_pool_entry entry,
                        uint32_t max_offset)
 {
+  if (_jit->overflow)
+    return 1;
+
   if (max_offset <= literal_pool_byte_size(_jit->pool)) {
     emit_literal_pool(_jit, GUARD_NEEDED);
-    ASSERT(_jit->pool->size == 0);
     return 0;
   }
 
@@ -1326,7 +1414,7 @@ remove_pending_literal(jit_state_t *_jit, jit_reloc_t src)
 }
 
 static void
-patch_pending_literal(jit_state_t *_jit, jit_reloc_t src, uint64_t value)
+patch_pending_literal(jit_state_t *_jit, jit_reloc_t src, uintptr_t value)
 {
   for (size_t i = _jit->pool->size; i--; ) {
     if (_jit->pool->entries[i].reloc.offset == src.offset) {
@@ -1353,10 +1441,12 @@ emit_literal_pool(jit_state_t *_jit, enum guard_pool guard)
 
   // FIXME: Could de-duplicate constants.
   for (size_t i = 0; i < _jit->pool->size; i++) {
-    // Align to 8-byte boundary without emitting pool.
+    // Align to word boundary without emitting pool.
     if (_jit->pc.w & 1) emit_u8(_jit, 0);
     if (_jit->pc.w & 2) emit_u16(_jit, 0);
-    if (_jit->pc.w & 4) emit_u32(_jit, 0);
+    if (sizeof(uintptr_t) > 4 && (_jit->pc.w & 4))
+      emit_u32(_jit, 0);
+    ASSERT((_jit->pc.w & (sizeof(uintptr_t) - 1)) == 0);
     struct jit_literal_pool_entry *entry = &_jit->pool->entries[i];
     uint8_t *loc = _jit->start + entry->reloc.offset;
     uint8_t *pc_base =
@@ -1367,18 +1457,18 @@ emit_literal_pool(jit_state_t *_jit, enum guard_pool guard)
     if (_jit->overflow)
       return;
 
-    switch (entry->reloc.kind) {
+    switch (entry->reloc.kind & JIT_RELOC_MASK) {
     case JIT_RELOC_JMP_WITH_VENEER:
-      patch_jmp_offset((uint32_t*) loc, diff);
-      emit_veneer(_jit, (void*) (uintptr_t) entry->value);
+      patch_veneer_jmp_offset((uint32_t*) loc, diff);
+      emit_veneer(_jit, (void*) entry->value);
       break;
     case JIT_RELOC_JCC_WITH_VENEER:
-      patch_jcc_offset((uint32_t*) loc, diff);
-      emit_veneer(_jit, (void*) (uintptr_t) entry->value);
+      patch_veneer_jcc_offset((uint32_t*) loc, diff);
+      emit_veneer(_jit, (void*) entry->value);
       break;
     case JIT_RELOC_LOAD_FROM_POOL:
       patch_load_from_pool_offset((uint32_t*) loc, diff);
-      emit_u64(_jit, entry->value);
+      emit_uintptr(_jit, entry->value);
       break;
     default:
       abort();
