@@ -1,21 +1,21 @@
-/* Copyright (C) 1995-1997, 1999-2001, 2003, 2004, 2006-2012, 2014, 2015
- *   Free Software Foundation, Inc.
- * 
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
- * the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
- * 02110-1301 USA
- */
+/* Copyright 1995-1997,1999-2001,2003-2004,2006-2012,2014-2020
+     Free Software Foundation, Inc.
+
+   This file is part of Guile.
+
+   Guile is free software: you can redistribute it and/or modify it
+   under the terms of the GNU Lesser General Public License as published
+   by the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
+
+   Guile is distributed in the hope that it will be useful, but WITHOUT
+   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+   License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with Guile.  If not, see
+   <https://www.gnu.org/licenses/>.  */
 
 
 
@@ -24,38 +24,46 @@
 # include <config.h>
 #endif
 
+#include <alloca.h>
+#include <c-ctype.h>
+#include <c-strcase.h>
 #include <stdio.h>
 #include <string.h>
-#include <unistd.h>
 #include <unicase.h>
 #include <unictype.h>
-#include <c-strcase.h>
-#include <c-ctype.h>
-#include <alloca.h>
+#include <unistd.h>
 
-#include "libguile/_scm.h"
-#include "libguile/bytevectors.h"
-#include "libguile/chars.h"
-#include "libguile/eval.h"
-#include "libguile/arrays.h"
-#include "libguile/bitvectors.h"
-#include "libguile/keywords.h"
-#include "libguile/alist.h"
-#include "libguile/srcprop.h"
-#include "libguile/hashtab.h"
-#include "libguile/hash.h"
-#include "libguile/ports.h"
-#include "libguile/ports-internal.h"
-#include "libguile/fports.h"
-#include "libguile/strings.h"
-#include "libguile/strports.h"
-#include "libguile/vectors.h"
-#include "libguile/validate.h"
-#include "libguile/srfi-4.h"
-#include "libguile/srfi-13.h"
+#include "alist.h"
+#include "arrays.h"
+#include "bitvectors.h"
+#include "boolean.h"
+#include "bytevectors.h"
+#include "chars.h"
+#include "eq.h"
+#include "eval.h"
+#include "fluids.h"
+#include "fports.h"
+#include "gsubr.h"
+#include "hash.h"
+#include "hashtab.h"
+#include "keywords.h"
+#include "modules.h"
+#include "numbers.h"
+#include "pairs.h"
+#include "ports-internal.h"
+#include "ports.h"
+#include "private-options.h"
+#include "procs.h"
+#include "srcprop.h"
+#include "srfi-13.h"
+#include "srfi-4.h"
+#include "strings.h"
+#include "strports.h"
+#include "symbols.h"
+#include "variable.h"
+#include "vectors.h"
 
-#include "libguile/read.h"
-#include "libguile/private-options.h"
+#include "read.h"
 
 
 
@@ -73,8 +81,6 @@ SCM_SYMBOL (sym_bracket_apply, "$bracket-apply$");
 
 scm_t_option scm_read_opts[] =
   {
-    { SCM_OPTION_BOOLEAN, "copy", 0,
-      "Copy source code expressions." },
     { SCM_OPTION_BOOLEAN, "positions", 1,
       "Record positions of source code expressions." },
     { SCM_OPTION_BOOLEAN, "case-insensitive", 0,
@@ -108,7 +114,6 @@ enum t_keyword_style
 struct t_read_opts
 {
   enum t_keyword_style keyword_style;
-  unsigned int copy_source_p        : 1;
   unsigned int record_positions_p   : 1;
   unsigned int case_insensitive_p   : 1;
   unsigned int r6rs_escapes_p       : 1;
@@ -142,21 +147,21 @@ scm_i_input_error (char const *function,
 {
   SCM fn = (scm_is_string (SCM_FILENAME(port))
 	    ? SCM_FILENAME(port)
-	    : scm_from_locale_string ("#<unknown port>"));
+	    : scm_from_utf8_string ("#<unknown port>"));
 
   SCM string_port = scm_open_output_string ();
   SCM string = SCM_EOL;
   scm_simple_format (string_port,
-		     scm_from_locale_string ("~A:~S:~S: ~A"),
+		     scm_from_utf8_string ("~A:~S:~S: ~A"),
 		     scm_list_4 (fn,
 				 scm_sum (scm_port_line (port), SCM_INUM1),
 				 scm_sum (scm_port_column (port), SCM_INUM1),
-				 scm_from_locale_string (message)));
+				 scm_from_utf8_string (message)));
     
   string = scm_get_output_string (string_port);
   scm_close_output_port (string_port);
-  scm_error_scm (scm_from_latin1_symbol ("read-error"),
-		 function? scm_from_locale_string (function) : SCM_BOOL_F,
+  scm_error_scm (scm_from_utf8_symbol ("read-error"),
+		 function? scm_from_utf8_string (function) : SCM_BOOL_F,
 		 string,
 		 arg,
 		 SCM_BOOL_F);
@@ -170,12 +175,7 @@ SCM_DEFINE (scm_read_options, "read-options-interface", 0, 1, 0,
 	    "@code{read-disable}, @code{read-set!} and @code{read-options}.")
 #define FUNC_NAME s_scm_read_options
 {
-  SCM ans = scm_options (setting,
-			 scm_read_opts,
-			 FUNC_NAME);
-  if (SCM_COPY_SOURCE_P)
-    SCM_RECORD_POSITIONS_P = 1;
-  return ans;
+  return scm_options (setting, scm_read_opts, FUNC_NAME);
 }
 #undef FUNC_NAME
 
@@ -405,20 +405,26 @@ flush_ws (SCM port, scm_t_read_opts *opts, const char *eoferr)
 
 static SCM scm_read_expression (SCM port, scm_t_read_opts *opts);
 static SCM scm_read_sharp (int chr, SCM port, scm_t_read_opts *opts,
-                           long line, int column);
+                           SCM line, SCM column);
 
 
 static SCM
 maybe_annotate_source (SCM x, SCM port, scm_t_read_opts *opts,
-                       long line, int column)
+                       SCM line, SCM column)
 {
-  /* This condition can be caused by a user calling
-     set-port-column!.  */
-  if (line < 0 || column < 0)
+  if ((SCM_I_INUMP (line) && SCM_I_INUM (line) < 0)
+      || (SCM_I_INUMP (column) && SCM_I_INUM (column) < 1))
+    /* This condition can be caused by a user calling
+       set-port-column!.  */
     return x;
 
   if (opts->record_positions_p)
-    scm_i_set_source_properties_x (x, line, column, SCM_FILENAME (port));
+    {
+      /* We always capture the column after one char of lookahead;
+         subtract off that lookahead value.  */
+      column = scm_oneminus (column);
+      scm_i_set_source_properties_x (x, line, column, SCM_FILENAME (port));
+    }
   return x;
 }
 
@@ -434,8 +440,8 @@ scm_read_sexp (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
                                    : ')'));
 
   /* Need to capture line and column numbers here. */
-  long line = scm_to_long (scm_port_line (port));
-  int column = scm_to_int (scm_port_column (port)) - 1;
+  SCM line = scm_port_line (port);
+  SCM column = scm_port_column (port);
 
   c = flush_ws (port, opts, FUNC_NAME);
   if (terminating_char == c)
@@ -612,8 +618,8 @@ scm_read_string_like_syntax (int chr, SCM port, scm_t_read_opts *opts)
   scm_t_wchar c, c_str[READER_STRING_BUFFER_SIZE];
 
   /* Need to capture line and column numbers here. */
-  long line = scm_to_long (scm_port_line (port));
-  int column = scm_to_int (scm_port_column (port)) - 1;
+  SCM line = scm_port_line (port);
+  SCM column = scm_port_column (port);
 
   while (chr != (c = scm_getc (port)))
     {
@@ -696,7 +702,7 @@ scm_read_string_like_syntax (int chr, SCM port, scm_t_read_opts *opts)
                 break;
             bad_escaped:
               scm_i_input_error (FUNC_NAME, port,
-                                 "illegal character in escape sequence: ~S",
+                                 "invalid character in escape sequence: ~S",
                                  scm_list_1 (SCM_MAKE_CHAR (c)));
             }
         }
@@ -739,8 +745,8 @@ scm_read_number (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
   size_t bytes_read;
 
   /* Need to capture line and column numbers here. */
-  long line = scm_to_long (scm_port_line (port));
-  int column = scm_to_int (scm_port_column (port)) - 1;
+  SCM line = scm_port_line (port);
+  SCM column = scm_port_column (port);
 
   scm_ungetc (chr, port);
   buffer = read_complete_token (port, opts, local_buffer, sizeof local_buffer,
@@ -866,8 +872,8 @@ static SCM
 scm_read_quote (int chr, SCM port, scm_t_read_opts *opts)
 {
   SCM p;
-  long line = scm_to_long (scm_port_line (port));
-  int column = scm_to_int (scm_port_column (port)) - 1;
+  SCM line = scm_port_line (port);
+  SCM column = scm_port_column (port);
 
   switch (chr)
     {
@@ -913,8 +919,8 @@ static SCM
 scm_read_syntax (int chr, SCM port, scm_t_read_opts *opts)
 {
   SCM p;
-  long line = scm_to_long (scm_port_line (port));
-  int column = scm_to_int (scm_port_column (port)) - 1;
+  SCM line = scm_port_line (port);
+  SCM column = scm_port_column (port);
 
   switch (chr)
     {
@@ -1092,7 +1098,7 @@ scm_read_character (scm_t_wchar chr, SCM port, scm_t_read_opts *opts)
   /* Ignore dotted circles, which may be used to keep combining characters from
      combining with the backslash in #\charname.  */
   if (cp == SCM_CODEPOINT_DOTTED_CIRCLE && charname_len == 2)
-    return SCM_MAKE_CHAR (scm_i_string_ref (charname, 1));
+    return scm_c_make_char (scm_i_string_ref (charname, 1));
 
   if (cp >= '0' && cp < '8')
     {
@@ -1169,7 +1175,7 @@ scm_read_keyword (int chr, SCM port, scm_t_read_opts *opts)
 
 static SCM
 scm_read_vector (int chr, SCM port, scm_t_read_opts *opts,
-                 long line, int column)
+                 SCM line, SCM column)
 {
   /* Note: We call `scm_read_sexp ()' rather than READER here in order to
      guarantee that it's going to do what we want.  After all, this is an
@@ -1214,7 +1220,7 @@ read_decimal_integer (SCM port, int c, ssize_t *resp)
 
    C is the first character read after the '#'. */
 static SCM
-scm_read_array (int c, SCM port, scm_t_read_opts *opts, long line, int column)
+scm_read_array (int c, SCM port, scm_t_read_opts *opts, SCM line, SCM column)
 {
   ssize_t rank;
   scm_t_wchar tag_buf[8];
@@ -1345,14 +1351,14 @@ scm_read_array (int c, SCM port, scm_t_read_opts *opts, long line, int column)
 
 static SCM
 scm_read_srfi4_vector (int chr, SCM port, scm_t_read_opts *opts,
-                       long line, int column)
+                       SCM line, SCM column)
 {
   return scm_read_array (chr, port, opts, line, column);
 }
 
 static SCM
 scm_read_bytevector (scm_t_wchar chr, SCM port, scm_t_read_opts *opts,
-                     long line, int column)
+                     SCM line, SCM column)
 {
   chr = scm_getc (port);
   if (chr != 'u')
@@ -1379,7 +1385,7 @@ scm_read_bytevector (scm_t_wchar chr, SCM port, scm_t_read_opts *opts,
 
 static SCM
 scm_read_guile_bit_vector (scm_t_wchar chr, SCM port, scm_t_read_opts *opts,
-                           long line, int column)
+                           SCM line, SCM column)
 {
   /* Read the `#*10101'-style read syntax for bit vectors in Guile.  This is
      terribly inefficient but who cares?  */
@@ -1396,7 +1402,7 @@ scm_read_guile_bit_vector (scm_t_wchar chr, SCM port, scm_t_read_opts *opts,
     scm_ungetc (chr, port);
 
   return maybe_annotate_source
-    (scm_bitvector (scm_reverse_x (s_bits, SCM_EOL)),
+    (scm_list_to_bitvector (scm_reverse_x (s_bits, SCM_EOL)),
      port, opts, line, column);
 }
 
@@ -1556,7 +1562,8 @@ scm_read_extended_symbol (scm_t_wchar chr, SCM port)
   size_t len = 0;
   SCM buf = scm_i_make_string (1024, NULL, 0);
 
-  buf = scm_i_string_start_writing (buf);
+  /* No need to scm_i_string_start_writing (), as the string isn't
+     visible to any other thread.  */
 
   while ((chr = scm_getc (port)) != EOF)
     {
@@ -1620,16 +1627,13 @@ scm_read_extended_symbol (scm_t_wchar chr, SCM port)
 	{
 	  SCM addy;
 
-	  scm_i_string_stop_writing ();
 	  addy = scm_i_make_string (1024, NULL, 0);
 	  buf = scm_string_append (scm_list_2 (buf, addy));
 	  len = 0;
-	  buf = scm_i_string_start_writing (buf);
 	}
     }
 
  done:
-  scm_i_string_stop_writing ();
   if (chr == EOF)
     scm_i_input_error ("scm_read_extended_symbol", port,
                        "end of file while reading symbol", SCM_EOL);
@@ -1649,8 +1653,8 @@ scm_read_sharp_extension (int chr, SCM port, scm_t_read_opts *opts)
   proc = scm_get_hash_procedure (chr);
   if (scm_is_true (scm_procedure_p (proc)))
     {
-      long line = scm_to_long (scm_port_line (port));
-      int column = scm_to_int (scm_port_column (port)) - 2;
+      SCM line = scm_port_line (port);
+      SCM column = scm_oneminus (scm_port_column (port));
       SCM got;
 
       got = scm_call_2 (proc, SCM_MAKE_CHAR (chr), port);
@@ -1669,7 +1673,7 @@ scm_read_sharp_extension (int chr, SCM port, scm_t_read_opts *opts)
    among the above token readers.   */
 static SCM
 scm_read_sharp (scm_t_wchar chr, SCM port, scm_t_read_opts *opts,
-                long line, int column)
+                SCM line, SCM column)
 #define FUNC_NAME "scm_lreadr"
 {
   SCM result;
@@ -1802,8 +1806,8 @@ read_inner_expression (SCM port, scm_t_read_opts *opts)
                  be part of an unescaped symbol.  We might as well do
                  something useful with it, so we adopt Kawa's convention:
                  [...] => ($bracket-list$ ...) */
-              long line = scm_to_long (scm_port_line (port));
-              int column = scm_to_int (scm_port_column (port)) - 1;
+              SCM line = scm_port_line (port);
+              SCM column = scm_port_column (port);
               return maybe_annotate_source
                 (scm_cons (sym_bracket_list, scm_read_sexp (chr, port, opts)),
                  port, opts, line, column);
@@ -1825,8 +1829,8 @@ read_inner_expression (SCM port, scm_t_read_opts *opts)
 	  return (scm_read_quote (chr, port, opts));
 	case '#':
 	  {
-            long line = scm_to_long (scm_port_line (port));
-            int column = scm_to_int (scm_port_column (port)) - 1;
+            SCM line = scm_port_line (port);
+            SCM column = scm_port_column (port);
 	    SCM result = scm_read_sharp (chr, port, opts, line, column);
 	    if (scm_is_eq (result, SCM_UNSPECIFIED))
 	      /* We read a comment or some such.  */
@@ -1874,8 +1878,8 @@ scm_read_expression (SCM port, scm_t_read_opts *opts)
     return read_inner_expression (port, opts);
   else
     {
-      long line = 0;
-      int column = 0;
+      SCM line = SCM_INUM0;
+      SCM column = SCM_INUM1;
       SCM expr;
 
       if (opts->record_positions_p)
@@ -1889,9 +1893,9 @@ scm_read_expression (SCM port, scm_t_read_opts *opts)
           int c = flush_ws (port, opts, (char *) NULL);
           if (c == EOF)
             return SCM_EOF_VAL;
+          line = scm_port_line (port);
+          column = scm_port_column (port);
           scm_ungetc (c, port);
-          line = scm_to_long (scm_port_line (port));
-          column = scm_to_int (scm_port_column (port));
         }
 
       expr = read_inner_expression (port, opts);
@@ -2080,7 +2084,7 @@ scm_i_scan_for_encoding (SCM port)
   scm_t_port *pt;
   SCM buf;
   char header[SCM_ENCODING_SEARCH_SIZE+1];
-  size_t bytes_read, encoding_length, i;
+  size_t cur, bytes_read, encoding_length, i;
   char *encoding = NULL;
   char *pos, *encoding_start;
   int in_comment;
@@ -2091,11 +2095,10 @@ scm_i_scan_for_encoding (SCM port)
   if (pt->rw_random)
     scm_flush (port);
 
-  if (scm_port_buffer_can_take (buf) == 0)
+  if (scm_port_buffer_can_take (buf, &cur) == 0)
     {
       /* We can use the read buffer, and thus avoid a seek. */
-      buf = scm_fill_input (port, 0);
-      bytes_read = scm_port_buffer_can_take (buf);
+      buf = scm_fill_input (port, 0, &cur, &bytes_read);
       if (bytes_read > SCM_ENCODING_SEARCH_SIZE)
         bytes_read = SCM_ENCODING_SEARCH_SIZE;
 
@@ -2103,24 +2106,20 @@ scm_i_scan_for_encoding (SCM port)
         /* An unbuffered port -- don't scan.  */
         return NULL;
 
-      memcpy (header, scm_port_buffer_take_pointer (buf), bytes_read);
+      memcpy (header, scm_port_buffer_take_pointer (buf, cur), bytes_read);
       header[bytes_read] = '\0';
     }
-  else
+  else if (pt->rw_random)
     {
-      /* Try to read some bytes and then seek back.  Not all ports
-         support seeking back; and indeed some file ports (like
-         /dev/urandom) will succeed on an lseek (fd, 0, SEEK_CUR)---the
-         check performed by SCM_FPORT_FDES---but fail to seek
-         backwards.  Hence this block comes second.  We prefer to use
-         the read buffer in-place.  */
-      if (SCM_FPORTP (port) && !SCM_FDES_RANDOM_P (SCM_FPORT_FDES (port)))
-        return NULL;
-
+      /* The port is seekable.  This is OK but grubbing in the read
+         buffer is better, so this case is just a fallback.  */
       bytes_read = scm_c_read (port, header, SCM_ENCODING_SEARCH_SIZE);
       header[bytes_read] = '\0';
       scm_seek (port, scm_from_int (0), scm_from_int (SEEK_SET));
     }
+  else
+    /* No input available and not seekable; scan fails.  */
+    return NULL;
 
   /* search past "coding[:=]" */
   pos = header;
@@ -2218,7 +2217,10 @@ SCM_DEFINE (scm_file_encoding, "file-encoding", 1, 0, 0,
     return SCM_BOOL_F;
   else
     {
-      s_enc = scm_string_upcase (scm_from_locale_string (enc));
+      /* It's not obvious what encoding to use here, but latin1 has the
+         advantage of never causing a decoding error, and a valid
+         encoding name should be ASCII anyway. */
+      s_enc = scm_string_upcase (scm_from_latin1_string (enc));
       return s_enc;
     }
 
@@ -2246,18 +2248,17 @@ SCM_DEFINE (scm_file_encoding, "file-encoding", 1, 0, 0,
 SCM_SYMBOL (sym_port_read_options, "port-read-options");
 
 /* Offsets of bit fields for each per-port override */
-#define READ_OPTION_COPY_SOURCE_P          0
-#define READ_OPTION_RECORD_POSITIONS_P     2
-#define READ_OPTION_CASE_INSENSITIVE_P     4
-#define READ_OPTION_KEYWORD_STYLE          6
-#define READ_OPTION_R6RS_ESCAPES_P         8
-#define READ_OPTION_SQUARE_BRACKETS_P     10
-#define READ_OPTION_HUNGRY_EOL_ESCAPES_P  12
-#define READ_OPTION_CURLY_INFIX_P         14
-#define READ_OPTION_R7RS_SYMBOLS_P        16
+#define READ_OPTION_RECORD_POSITIONS_P     0
+#define READ_OPTION_CASE_INSENSITIVE_P     2
+#define READ_OPTION_KEYWORD_STYLE          4
+#define READ_OPTION_R6RS_ESCAPES_P         6
+#define READ_OPTION_SQUARE_BRACKETS_P      8
+#define READ_OPTION_HUNGRY_EOL_ESCAPES_P  10
+#define READ_OPTION_CURLY_INFIX_P         12
+#define READ_OPTION_R7RS_SYMBOLS_P        14
 
 /* The total width in bits of the per-port overrides */
-#define READ_OPTIONS_NUM_BITS             18
+#define READ_OPTIONS_NUM_BITS             16
 
 #define READ_OPTIONS_INHERIT_ALL  ((1UL << READ_OPTIONS_NUM_BITS) - 1)
 #define READ_OPTIONS_MAX_VALUE    READ_OPTIONS_INHERIT_ALL
@@ -2373,7 +2374,6 @@ init_read_options (SCM port, scm_t_read_opts *opts)
     }                                                                   \
   while (0)
 
-  RESOLVE_BOOLEAN_OPTION (COPY_SOURCE_P,        copy_source_p);
   RESOLVE_BOOLEAN_OPTION (RECORD_POSITIONS_P,   record_positions_p);
   RESOLVE_BOOLEAN_OPTION (CASE_INSENSITIVE_P,   case_insensitive_p);
   RESOLVE_BOOLEAN_OPTION (R6RS_ESCAPES_P,       r6rs_escapes_p);
@@ -2398,11 +2398,5 @@ scm_init_read ()
     SCM_VARIABLE_LOC (scm_c_define ("%read-hash-procedures", read_hash_procs));
 
   scm_init_opts (scm_read_options, scm_read_opts);
-#include "libguile/read.x"
+#include "read.x"
 }
-
-/*
-  Local Variables:
-  c-file-style: "gnu"
-  End:
-*/
