@@ -1,47 +1,47 @@
-/* Copyright (C) 1995,1996,1998,1999,2000,2001, 2002, 2003, 2006, 2009, 2010, 2011, 2013 Free Software Foundation, Inc.
- * 
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
- * the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
- * 02110-1301 USA
- */
+/* Copyright 1995-1996,1998-2003,2006,2009-2011,2013,2018
+     Free Software Foundation, Inc.
+
+   This file is part of Guile.
+
+   Guile is free software: you can redistribute it and/or modify it
+   under the terms of the GNU Lesser General Public License as published
+   by the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
+
+   Guile is distributed in the hope that it will be useful, but WITHOUT
+   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+   License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with Guile.  If not, see
+   <https://www.gnu.org/licenses/>.  */
 
 
 
+
 #ifdef HAVE_CONFIG_H
-#  include <config.h>
+# include <config.h>
 #endif
 
 #include <assert.h>
-#include <stdio.h>
 #include <errno.h>
-
-#include "libguile/_scm.h"
-#include "libguile/eval.h"
-#include "libguile/chars.h"
-#include "libguile/ports.h"
-#include "libguile/ports-internal.h"
-#include "libguile/fports.h"
-#include "libguile/root.h"
-#include "libguile/strings.h"
-#include "libguile/vectors.h"
-
-#include "libguile/validate.h"
-#include "libguile/vports.h"
-
-#ifdef HAVE_STRING_H
+#include <stdio.h>
 #include <string.h>
-#endif
+
+#include "boolean.h"
+#include "chars.h"
+#include "eval.h"
+#include "fports.h"
+#include "gsubr.h"
+#include "numbers.h"
+#include "ports-internal.h"
+#include "ports.h"
+#include "strings.h"
+#include "vectors.h"
+
+#include "vports.h"
+
 
 
 
@@ -49,7 +49,6 @@
  * 
  */
 
-
 static scm_t_port_type *scm_soft_port_type;
 
 #define ENCODE_BUF_SIZE 10
@@ -61,7 +60,7 @@ struct soft_port {
   SCM read_char;
   SCM close;
   SCM input_waiting;
-  scm_t_uint8 encode_buf[ENCODE_BUF_SIZE];
+  uint8_t encode_buf[ENCODE_BUF_SIZE];
   size_t encode_cur;
   size_t encode_end;
 };
@@ -247,11 +246,5 @@ scm_init_vports ()
 {
   scm_soft_port_type = scm_make_sfptob ();
 
-#include "libguile/vports.x"
+#include "vports.x"
 }
-
-/*
-  Local Variables:
-  c-file-style: "gnu"
-  End:
-*/
