@@ -1,6 +1,6 @@
 ;;; Continuation-passing style (CPS) intermediate language (IL)
 
-;; Copyright (C) 2013-2015, 2017-2019 Free Software Foundation, Inc.
+;; Copyright (C) 2013-2015, 2017-2020 Free Software Foundation, Inc.
 
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
@@ -80,6 +80,8 @@
           (ref* args))))
       (($ $kargs _ _ ($ $branch kf kt src op param args))
        (ref* args))
+      (($ $kargs _ _ ($ $switch kf kt* src arg))
+       (ref arg))
       (($ $kargs _ _ ($ $prompt k kh src escape? tag))
        (ref tag))
       (($ $kargs _ _ ($ $throw src op param args))
@@ -149,6 +151,9 @@
               (($ $kargs names syms ($ $branch kf kt src op param args))
                ($kargs names syms
                  ($branch (subst kf) (subst kt) src op param args)))
+              (($ $kargs names syms ($ $switch kf kt* src arg))
+               ($kargs names syms
+                 ($switch (subst kf) (map subst kt*) src arg)))
               (($ $kargs names syms ($ $prompt k kh src escape? tag))
                ($kargs names syms
                  ($prompt (subst k) (subst kh) src escape? tag)))
@@ -175,38 +180,12 @@
               (_ ,cont))))
      conts)))
 
-(define (compute-singly-referenced-labels conts body)
-  (define (add-ref label single multiple)
-    (define (ref k single multiple)
-      (if (intset-ref single k)
-          (values single (intset-add! multiple k))
-          (values (intset-add! single k) multiple)))
-    (define (ref0) (values single multiple))
-    (define (ref1 k) (ref k single multiple))
-    (define (ref2 k k*)
-      (if k*
-          (let-values (((single multiple) (ref k single multiple)))
-            (ref k* single multiple))
-          (ref1 k)))
-    (match (intmap-ref conts label)
-      (($ $kreceive arity k) (ref1 k))
-      (($ $kfun src meta self ktail kclause) (ref2 ktail kclause))
-      (($ $ktail) (ref0))
-      (($ $kclause arity kbody kalt) (ref2 kbody kalt))
-      (($ $kargs names syms ($ $continue k)) (ref1 k))
-      (($ $kargs names syms ($ $branch kf kt)) (ref2 kf kt))
-      (($ $kargs names syms ($ $prompt k kh)) (ref2 k kh))
-      (($ $kargs names syms ($ $throw)) (ref0))))
-  (let*-values (((single multiple) (values empty-intset empty-intset))
-                ((single multiple) (intset-fold add-ref body single multiple)))
-    (intset-subtract (persistent-intset single)
-                     (persistent-intset multiple))))
-
 (define (compute-beta-reductions conts kfun)
   (define (visit-fun kfun body beta)
-    (let ((single (compute-singly-referenced-labels conts body)))
-      (define (visit-cont label beta)
-        (match (intmap-ref conts label)
+    (let* ((conts (intmap-select conts body))
+           (single (compute-singly-referenced-labels conts)))
+      (define (visit-cont label cont beta)
+        (match cont
           ;; A continuation's body can be inlined in place of a $values
           ;; expression if the continuation is a $kargs.  It should only
           ;; be inlined if it is used only once, and not recursively.
@@ -218,7 +197,7 @@
                                      (_ #f)))))
           (_
            beta)))
-      (intset-fold visit-cont body beta)))
+      (intmap-fold visit-cont conts beta)))
   (persistent-intset
    (intmap-fold visit-fun
                 (compute-reachable-functions conts kfun)
@@ -266,6 +245,8 @@
                    ($values ,(map subst args))))))
             (($ $branch kf kt src op param args)
              ($branch kf kt src op param ,(map subst args)))
+            (($ $switch kf kt* src arg)
+             ($switch kf kt* src (subst arg)))
             (($ $prompt k kh src escape? tag)
              ($prompt k kh src escape? (subst tag)))
             (($ $throw src op param args)
