@@ -1,6 +1,6 @@
 ;;; Guile bytecode disassembler
 
-;;; Copyright (C) 2001, 2009, 2010, 2012, 2013, 2014, 2015 Free Software Foundation, Inc.
+;;; Copyright (C) 2001, 2009-2010, 2012-2015, 2017-2020 Free Software Foundation, Inc.
 ;;;
 ;;; This library is free software; you can redistribute it and/or
 ;;; modify it under the terms of the GNU Lesser General Public
@@ -24,6 +24,7 @@
   #:use-module (system vm debug)
   #:use-module (system vm program)
   #:use-module (system vm loader)
+  #:use-module (system base types internal)
   #:use-module (system foreign)
   #:use-module (rnrs bytevectors)
   #:use-module (ice-9 format)
@@ -42,27 +43,6 @@
             instruction-stack-size-after
             instruction-slot-clobbers))
 
-(define-syntax-rule (u32-ref buf n)
-  (bytevector-u32-native-ref buf (* n 4)))
-
-(define-syntax-rule (s32-ref buf n)
-  (bytevector-s32-native-ref buf (* n 4)))
-
-(define-syntax visit-opcodes
-  (lambda (x)
-    (syntax-case x ()
-      ((visit-opcodes macro arg ...)
-       (with-syntax (((inst ...)
-                      (map (lambda (x) (datum->syntax #'macro x))
-                           (instruction-list))))
-         #'(begin
-             (macro arg ... . inst)
-             ...))))))
-
-(eval-when (expand compile load eval)
-  (define (id-append ctx a b)
-    (datum->syntax ctx (symbol-append (syntax->datum a) (syntax->datum b)))))
-
 (define (unpack-scm n)
   (pointer->scm (make-pointer n)))
 
@@ -71,13 +51,41 @@
       s
       (- s (ash 1 24))))
 
+(define (unpack-s12 s)
+  (if (zero? (logand s (ash 1 11)))
+      s
+      (- s (ash 1 12))))
+
 (define (unpack-s32 s)
   (if (zero? (logand s (ash 1 31)))
       s
       (- s (ash 1 32))))
 
-(define-syntax disassembler
-  (lambda (x)
+(eval-when (expand)
+  (define-syntax-rule (u32-ref buf n)
+    (bytevector-u32-native-ref buf (* n 4)))
+
+  (define-syntax-rule (s32-ref buf n)
+    (bytevector-s32-native-ref buf (* n 4)))
+
+  (define-syntax-rule (define-op-handlers handlers make-handler)
+    (define handlers
+      (let ((handlers (make-vector 256 #f)))
+        (define-syntax init-handlers
+          (lambda (stx)
+            #`(begin
+                #,@(filter-map
+                    (match-lambda
+                     ((name opcode kind . word-types)
+                      (match (make-handler name kind word-types)
+                        (#f #f)
+                        (init #`(vector-set! handlers #,opcode #,init)))))
+                    (instruction-list)))))
+        (init-handlers)
+        handlers))))
+
+(define-op-handlers disassemblers
+  (lambda (name kind word-types)
     (define (parse-first-word word type)
       (with-syntax ((word word))
         (case type
@@ -87,7 +95,7 @@
            #'((ash word -8)))
           ((X8_L24)
            #'((unpack-s24 (ash word -8))))
-          ((X8_S8_I16)
+          ((X8_S8_I16 X8_S8_ZI16)
            #'((logand (ash word -8) #xff)
               (ash word -16)))
           ((X8_S12_S12
@@ -96,6 +104,9 @@
             X8_F12_F12)
            #'((logand (ash word -8) #xfff)
               (ash word -20)))
+          ((X8_S12_Z12)
+           #'((logand (ash word -8) #xfff)
+              (unpack-s12 (ash word -20))))
           ((X8_S8_S8_S8
             X8_S8_S8_C8
             X8_S8_C8_S8)
@@ -105,78 +116,107 @@
           (else
            (error "bad head kind" type)))))
 
-    (define (parse-tail-word word type)
-      (with-syntax ((word word))
+    (define (parse-tail-word word type n)
+      (with-syntax ((word word) (n n))
         (case type
           ((C32 I32 A32 B32 AU32 BU32 AS32 BS32 AF32 BF32)
-           #'(word))
+           #'(1 word))
           ((N32 R32 L32 LO32)
-           #'((unpack-s32 word)))
-          ((C8_C24)
-           #'((logand word #xff)
+           #'(1 (unpack-s32 word)))
+          ((C8_C24 C8_S24)
+           #'(1
+              (logand word #xff)
               (ash word -8)))
+          ((C16_C16)
+           #'(1
+              (logand word #xffff)
+              (ash word -16)))
           ((B1_C7_L24)
-           #'((not (zero? (logand word #x1)))
+           #'(1
+              (not (zero? (logand word #x1)))
               (logand (ash word -1) #x7f)
               (unpack-s24 (ash word -8))))
           ((B1_X7_S24 B1_X7_F24 B1_X7_C24)
-           #'((not (zero? (logand word #x1)))
+           #'(1
+              (not (zero? (logand word #x1)))
               (ash word -8)))
           ((B1_X7_L24)
-           #'((not (zero? (logand word #x1)))
+           #'(1
+              (not (zero? (logand word #x1)))
               (unpack-s24 (ash word -8))))
           ((B1_X31)
-           #'((not (zero? (logand word #x1)))))
+           #'(1 (not (zero? (logand word #x1)))))
           ((X8_S24 X8_F24 X8_C24)
-           #'((ash word -8)))
+           #'(1 (ash word -8)))
           ((X8_L24)
-           #'((unpack-s24 (ash word -8))))
+           #'(1 (unpack-s24 (ash word -8))))
+          ((V32_X8_L24)
+           #'((+ 1 word)
+              (let ((v (make-vector word))
+                    (base (+ offset n 1)))
+                (let lp ((i 0))
+                  (when (< i word)
+                    (vector-set! v i
+                                 (unpack-s24 (ash (u32-ref buf (+ base i)) -8)))
+                    (lp (1+ i))))
+                v)))
           (else
            (error "bad tail kind" type)))))
 
-    (syntax-case x ()
-      ((_ name opcode word0 word* ...)
-       (let ((vars (generate-temporaries #'(word* ...))))
-         (with-syntax (((word* ...) vars)
-                       ((n ...) (map 1+ (iota (length #'(word* ...)))))
+    (match word-types
+      ((first-word . tail-words)
+       (let ((vars (generate-temporaries tail-words))
+             (word-offsets (map 1+ (iota (length tail-words)))))
+         (with-syntax ((name (datum->syntax #'nowhere name))
+                       ((word* ...) vars)
+                       ((n ...) word-offsets)
                        ((asm ...)
-                        (parse-first-word #'first (syntax->datum #'word0)))
-                       (((asm* ...) ...)
-                        (map (lambda (word type)
-                               (parse-tail-word word type))
-                             vars
-                             (syntax->datum #'(word* ...)))))
+                        (parse-first-word #'first first-word))
+                       (((len asm* ...) ...)
+                        (map parse-tail-word vars tail-words word-offsets)))
            #'(lambda (buf offset first)
                (let ((word* (u32-ref buf (+ offset n)))
                      ...)
-                 (values (+ 1 (length '(word* ...)))
+                 (values (+ 1 len ...)
                          (list 'name asm ... asm* ... ...))))))))))
 
-(define (disasm-invalid buf offset first)
-  (error "bad instruction" (logand first #xff) first buf offset))
-
-(define disassemblers (make-vector 256 disasm-invalid))
-
-(define-syntax define-disassembler
-  (lambda (x)
-    (syntax-case x ()
-      ((_ name opcode kind arg ...)
-       (with-syntax ((parse (id-append #'name #'parse- #'name)))
-         #'(let ((parse (disassembler name opcode arg ...)))
-             (vector-set! disassemblers opcode parse)))))))
-
-(visit-opcodes define-disassembler)
-
 ;; -> len list
 (define (disassemble-one buf offset)
   (let ((first (u32-ref buf offset)))
-    ((vector-ref disassemblers (logand first #xff)) buf offset first)))
+    (match (vector-ref disassemblers (logand first #xff))
+      (#f (error "bad instruction" (logand first #xff) first buf offset))
+      (disassemble (disassemble buf offset first)))))
 
 (define (u32-offset->addr offset context)
   "Given an offset into an image in 32-bit units, return the absolute
 address of that offset."
   (+ (debug-context-base context) (* offset 4)))
 
+(define immediate-tag-annotations '())
+(define-syntax-rule (define-immediate-tag-annotation name pred mask tag)
+  (set! immediate-tag-annotations
+        (cons `((,mask ,tag)
+                ,(cond
+                  ('pred => symbol->string)
+                  (else (string-append "eq-" (symbol->string 'name) "?"))))
+              immediate-tag-annotations)))
+(visit-immediate-tags define-immediate-tag-annotation)
+
+(define heap-tag-annotations '())
+(define-syntax-rule (define-heap-tag-annotation name pred mask tag)
+  (set! heap-tag-annotations
+        (cons `((,mask ,tag) ,(symbol->string 'pred)) heap-tag-annotations)))
+
+(visit-heap-tags define-heap-tag-annotation)
+
+(define (sign-extended-immediate uimm n)
+  (unpack-scm
+   (if (>= uimm (ash 1 (- n 1)))
+       (let ((word-bits (* (sizeof '*) 8))) ; FIXME
+         (logand (1- (ash 1 word-bits))
+                 (- uimm (ash 1 n))))
+       uimm)))
+
 (define (code-annotation code len offset start labels context push-addr!)
   ;; FIXME: Print names for register loads and stores that correspond to
   ;; access to named locals.
@@ -190,32 +230,17 @@ address of that offset."
        (dereference-pointer (make-pointer addr)))))
 
   (match code
-    (((or 'br
-          'br-if-nargs-ne 'br-if-nargs-lt 'br-if-nargs-gt
-          'br-if-true 'br-if-null 'br-if-nil 'br-if-pair 'br-if-struct
-          'br-if-char 'br-if-eq 'br-if-eqv
-          'br-if-= 'br-if-< 'br-if-<= 'br-if-> 'br-if->=
-          'br-if-u64-= 'br-if-u64-< 'br-if-u64-<=
-          'br-if-u64-<-scm 'br-if-u64-<=-scm 'br-if-u64-=-scm
-          'br-if-u64->-scm 'br-if-u64->=-scm
-          'br-if-logtest) _ ... target)
+    (((or 'j 'je 'jl 'jge 'jne 'jnl 'jnge) target)
      (list "-> ~A" (vector-ref labels (- (+ offset target) start))))
-    (('br-if-tc7 slot invert? tc7 target)
-     (list "~A -> ~A"
-           (let ((tag (case tc7
-                        ((5) "symbol?")
-                        ((7) "variable?")
-                        ((13) "vector?")
-                        ((15) "string?")
-                        ((53) "keyword?")
-                        ((77) "bytevector?")
-                        ((95) "bitvector?")
-                        (else (number->string tc7)))))
-             (if invert? (string-append "not " tag) tag))
-           (vector-ref labels (- (+ offset target) start))))
+    (('immediate-tag=? _ mask tag)
+     (assoc-ref immediate-tag-annotations (list mask tag)))
+    (('heap-tag=? _ mask tag)
+     (assoc-ref heap-tag-annotations (list mask tag)))
     (('prompt tag escape-only? proc-slot handler)
      ;; The H is for handler.
      (list "H -> ~A" (vector-ref labels (- (+ offset handler) start))))
+    (((or 'make-immediate 'eq-immediate?) _ imm)
+     (list "~S" (sign-extended-immediate imm 16)))
     (((or 'make-short-immediate 'make-long-immediate) _ imm)
      (list "~S" (unpack-scm imm)))
     (('make-long-long-immediate _ high low)
@@ -223,18 +248,14 @@ address of that offset."
     (('assert-nargs-ee/locals nargs locals)
      ;; The nargs includes the procedure.
      (list "~a slot~:p (~a arg~:p)" (+ locals nargs) (1- nargs)))
+    (('bind-optionals nargs)
+     (list "~a args~:p" (1- nargs)))
     (('alloc-frame nlocals)
      (list "~a slot~:p" nlocals))
     (('reset-frame nlocals)
      (list "~a slot~:p" nlocals))
-    (('return-values nlocals)
-     (if (zero? nlocals)
-         (list "all values")
-         (list "~a value~:p" (1- nlocals))))
     (('bind-rest dst)
      (list "~a slot~:p" (1+ dst)))
-    (('tail-call nargs proc)
-     (list "~a arg~:p" nargs))
     (('make-closure dst target nfree)
      (let* ((addr (u32-offset->addr (+ offset target) context))
             (pdi (find-program-debug-info addr context))
@@ -242,6 +263,13 @@ address of that offset."
                       "anonymous procedure")))
        (push-addr! addr name)
        (list "~A at #x~X (~A free var~:p)" name addr nfree)))
+    (('load-label dst src)
+     (let* ((addr (u32-offset->addr (+ offset src) context))
+            (pdi (find-program-debug-info addr context))
+            (name (or (and pdi (program-debug-info-name pdi))
+                      "anonymous procedure")))
+       (push-addr! addr name)
+       (list "~A at #x~X" name addr)))
     (('call-label closure nlocals target)
      (let* ((addr (u32-offset->addr (+ offset target) context))
             (pdi (find-program-debug-info addr context))
@@ -249,7 +277,7 @@ address of that offset."
                       "anonymous procedure")))
        (push-addr! addr name)
        (list "~A at #x~X" name addr)))
-    (('tail-call-label nlocals target)
+    (('tail-call-label target)
      (let* ((addr (u32-offset->addr (+ offset target) context))
             (pdi (find-program-debug-info addr context))
             (name (or (and pdi (program-debug-info-name pdi))
@@ -261,22 +289,14 @@ address of that offset."
        (when (program? val)
          (push-addr! (program-code val) val))
        (list "~@Y" val)))
+    (((or 'throw/value 'throw/value+data) dst target)
+     (list "~@Y" (reference-scm target)))
     (('builtin-ref dst idx)
      (list "~A" (builtin-index->name idx)))
     (((or 'static-ref 'static-set!) _ target)
      (list "~@Y" (dereference-scm target)))
-    (((or 'free-ref 'free-set!) _ _ index)
-     (list "free var ~a" index))
     (('resolve-module dst name public)
      (list "~a" (if (zero? public) "private" "public")))
-    (('toplevel-box _ var-offset mod-offset sym-offset bound?)
-     (list "`~A'~A" (dereference-scm sym-offset)
-           (if bound? "" " (maybe unbound)")))
-    (('module-box _ var-offset mod-name-offset sym-offset bound?)
-     (let ((mod-name (reference-scm mod-name-offset)))
-       (list "`(~A ~A ~A)'~A" (if (car mod-name) '@ '@@) (cdr mod-name)
-             (dereference-scm sym-offset)
-             (if bound? "" " (maybe unbound)"))))
     (('load-typed-array dst type shape target len)
      (let ((addr (u32-offset->addr (+ offset target) context)))
        (list "~a bytes from #x~X" len addr)))
@@ -295,21 +315,22 @@ address of that offset."
             (match elt
               ((inst arg ...)
                (case inst
-                 ((br
-                   br-if-nargs-ne br-if-nargs-lt br-if-nargs-gt
-                   br-if-true br-if-null br-if-nil br-if-pair br-if-struct
-                   br-if-char br-if-tc7 br-if-eq br-if-eqv
-                   br-if-= br-if-< br-if-<= br-if-> br-if->= br-if-logtest
-                   br-if-u64-= br-if-u64-< br-if-u64-<=
-                   br-if-u64-<-scm br-if-u64-<=-scm br-if-u64-=-scm
-                   br-if-u64->-scm br-if-u64->=-scm)
+                 ((j je jl jge jne jnl jnge)
                   (match arg
                     ((_ ... target)
                      (add-label! (+ offset target) "L"))))
                  ((prompt)
                   (match arg
                     ((_ ... target)
-                     (add-label! (+ offset target) "H")))))))
+                     (add-label! (+ offset target) "H"))))
+                 ((jtable)
+                  (match arg
+                    ((_ ... targets)
+                     (let ((len (vector-length targets)))
+                       (let lp ((i 0))
+                         (when (< i len)
+                           (add-label! (+ offset (vector-ref targets i)) "L")
+                           (lp (1+ i)))))))))))
             (lp (+ offset len))))))
     (let lp ((offset start) (n 1))
       (when (< offset end)
@@ -410,26 +431,14 @@ address of that offset."
        `(make-closure ,dst
                       ,(u32-offset->addr (+ offset target) context)
                       ,nfree))
+      (('load-label dst src)
+       `(load-label ,dst ,(u32-offset->addr (+ offset src) context)))
       (('make-non-immediate dst target)
        `(make-non-immediate ,dst ,(reference-scm target)))
       (('builtin-ref dst idx)
        `(builtin-ref ,dst ,(builtin-index->name idx)))
       (((or 'static-ref 'static-set!) dst target)
        `(,(car code) ,dst ,(dereference-scm target)))
-      (('toplevel-box dst var-offset mod-offset sym-offset bound?)
-       `(toplevel-box ,dst
-                      ,(dereference-scm var-offset)
-                      ,(dereference-scm mod-offset)
-                      ,(dereference-scm sym-offset)
-                      ,bound?))
-      (('module-box dst var-offset mod-name-offset sym-offset bound?)
-       (let ((mod-name (reference-scm mod-name-offset)))
-         `(module-box ,dst
-                      ,(dereference-scm var-offset)
-                      ,(car mod-name)
-                      ,(cdr mod-name)
-                      ,(dereference-scm sym-offset)
-                      ,bound?)))
       (_ code)))
   (let lp ((offset start) (seed seed))
     (cond
@@ -478,10 +487,10 @@ address of that offset."
          (display "\n\n" port)))))
   (values))
 
-(define (disassemble-file file)
+(define* (disassemble-file file #:optional (port (current-output-port)))
   (let* ((thunk (load-thunk-from-file file))
          (elf (find-mapped-elf-image (program-code thunk))))
-    (disassemble-image elf)))
+    (disassemble-image elf port)))
 
 (define-syntax instruction-lengths-vector
   (lambda (x)
@@ -489,15 +498,27 @@ address of that offset."
       ((_)
        (let ((lengths (make-vector 256 #f)))
          (for-each (match-lambda
+                    ((name opcode kind word ... 'V32_X8_L24)
+                     ;; Indicate variable-length instruction by setting
+                     ;; statically known length to 0.
+                     (vector-set! lengths opcode 0))
                     ((name opcode kind words ...)
                      (vector-set! lengths opcode (* 4 (length words)))))
                    (instruction-list))
          (datum->syntax x lengths))))))
 
 (define (instruction-length code pos)
+  (unless (zero? (modulo pos 4))
+    (error "invalid pos"))
   (let ((opcode (logand (bytevector-u32-native-ref code pos) #xff)))
-    (or (vector-ref (instruction-lengths-vector) opcode)
-        (error "Unknown opcode" opcode))))
+    (match (vector-ref (instruction-lengths-vector) opcode)
+      (#f (error "Unknown opcode" opcode))
+      (0 (call-with-values (lambda ()
+                             (let ((offset (/ pos 4)))
+                               (disassemble-one code offset)))
+           (lambda (u32-len disasm)
+             (* u32-len 4))))
+      (len len))))
 
 (define-syntax static-opcode-set
   (lambda (x)
@@ -512,144 +533,139 @@ address of that offset."
       ((static-opcode-set inst ...)
        (let ((bv (make-bitvector 256 #f)))
          (for-each (lambda (inst)
-                     (bitvector-set! bv (instruction-opcode inst) #t))
+                     (bitvector-set-bit! bv (instruction-opcode inst)))
                    (syntax->datum #'(inst ...)))
          (datum->syntax #'static-opcode-set bv))))))
 
 (define (instruction-has-fallthrough? code pos)
   (define non-fallthrough-set
     (static-opcode-set halt
-                       tail-call tail-call-label tail-call/shuffle
+                       throw throw/value throw/value+data
+                       tail-call tail-call-label
                        return-values
                        subr-call foreign-call continuation-call
-                       tail-apply
-                       br))
+                       j jtable))
   (let ((opcode (logand (bytevector-u32-native-ref code pos) #xff)))
-    (not (bitvector-ref non-fallthrough-set opcode))))
+    (bitvector-bit-clear? non-fallthrough-set opcode)))
 
-(define-syntax define-jump-parser
-  (lambda (x)
-    (syntax-case x ()
-      ((_ name opcode kind word0 word* ...)
-       (let ((symname (syntax->datum #'name)))
-         (if (or (memq symname '(br prompt))
-                 (string-prefix? "br-" (symbol->string symname)))
-             (let ((offset (* 4 (length #'(word* ...)))))
-               #`(vector-set!
-                  jump-parsers
-                  opcode
-                  (lambda (code pos)
-                    (let ((target
-                           (bytevector-s32-native-ref code (+ pos #,offset))))
-                      ;; Assume that the target is in the last word, as
-                      ;; an L24 in the high bits.
-                      (list (* 4 (ash target -8)))))))
-             #'(begin)))))))
+(define (word-offset->byte-offset n)
+  (* n 4))
 
-(define jump-parsers (make-vector 256 (lambda (code pos) '())))
-(visit-opcodes define-jump-parser)
+(define-op-handlers jump-parsers
+  (lambda (op kind word-types)
+    (case op
+      ((prompt j je jl jge jne jnl jnge)
+       #'(lambda (code pos)
+           (call-with-values (lambda () (disassemble-one code (/ pos 4)))
+             (lambda (len disasm)
+               (match disasm
+                 ;; Assume that the target is in the last word, as a
+                 ;; word offset.
+                 ((_ ___ target) (list (word-offset->byte-offset target))))))))
+      ((jtable)
+       #'(lambda (code pos)
+           (call-with-values (lambda () (disassemble-one code (/ pos 4)))
+             (lambda (len disasm)
+               (match disasm
+                 ;; Assume that the target is in the last word, as a
+                 ;; vector of word offsets.
+                 ((_ ___ targets)
+                  (map word-offset->byte-offset (vector->list targets))))))))
+      (else #f))))
 
 (define (instruction-relative-jump-targets code pos)
   (let ((opcode (logand (bytevector-u32-native-ref code pos) #xff)))
-    ((vector-ref jump-parsers opcode) code pos)))
+    (match (vector-ref jump-parsers opcode)
+      (#f '())
+      (proc (proc code pos)))))
 
-(define-syntax define-stack-effect-parser
-  (lambda (x)
-    (define (stack-effect-parser name)
-      (case name
-        ((push)
-         #'(lambda (code pos size) (+ size 1)))
-        ((pop)
-         #'(lambda (code pos size) (- size 1)))
-        ((drop)
-         #'(lambda (code pos size)
-             (let ((count (ash (bytevector-u32-native-ref code pos) -8)))
-               (- size count))))
-        ((alloc-frame reset-frame)
-         #'(lambda (code pos size)
-             (let ((nlocals (ash (bytevector-u32-native-ref code pos) -8)))
-               nlocals)))
-        ((receive)
-         #'(lambda (code pos size)
-             (let ((nlocals (ash (bytevector-u32-native-ref code (+ pos 4))
-                                 -8)))
-               nlocals)))
-        ((bind-kwargs)
-         #'(lambda (code pos size)
-             (let ((ntotal (ash (bytevector-u32-native-ref code (+ pos 8)) -8)))
-               ntotal)))
-        ((bind-rest)
-         #'(lambda (code pos size)
-             (let ((dst (ash (bytevector-u32-native-ref code pos) -8)))
-               (+ dst 1))))
-        ((assert-nargs-ee/locals)
-         #'(lambda (code pos size)
-             (let ((nargs (logand (ash (bytevector-u32-native-ref code pos) -8)
-                                  #xfff))
-                   (nlocals (ash (bytevector-u32-native-ref code pos) -20)))
-               (+ nargs nlocals))))
-        ((call call-label)
-         #'(lambda (code pos size) #f))
-        ((tail-call tail-call-label tail-call/shuffle tail-apply)
-         #'(lambda (code pos size) #f))
-        (else
-         #f)))
-    (syntax-case x ()
-      ((_ name opcode kind word0 word* ...)
-       (let ((parser (stack-effect-parser (syntax->datum #'name))))
-         (if parser
-             #`(vector-set! stack-effect-parsers opcode #,parser)
-             #'(begin)))))))
-
-(define stack-effect-parsers (make-vector 256 (lambda (code pos size) size)))
-(visit-opcodes define-stack-effect-parser)
+(define-op-handlers stack-effect-parsers
+  (lambda (name kind word-types)
+    (case name
+      ((push)
+       #'(lambda (code pos size) (and size (+ size 1))))
+      ((pop)
+       #'(lambda (code pos size) (and size (- size 1))))
+      ((drop)
+       #'(lambda (code pos size)
+           (let ((count (ash (bytevector-u32-native-ref code pos) -8)))
+             (and size (- size count)))))
+      ((alloc-frame reset-frame bind-optionals)
+       #'(lambda (code pos size)
+           (let ((nlocals (ash (bytevector-u32-native-ref code pos) -8)))
+             nlocals)))
+      ((receive)
+       #'(lambda (code pos size)
+           (let ((nlocals (ash (bytevector-u32-native-ref code (+ pos 4))
+                               -8)))
+             nlocals)))
+      ((bind-kwargs)
+       #'(lambda (code pos size)
+           (let ((ntotal (ash (bytevector-u32-native-ref code (+ pos 8)) -8)))
+             ntotal)))
+      ((bind-rest)
+       #'(lambda (code pos size)
+           (let ((dst (ash (bytevector-u32-native-ref code pos) -8)))
+             (+ dst 1))))
+      ((assert-nargs-ee/locals)
+       #'(lambda (code pos size)
+           (let ((nargs (logand (ash (bytevector-u32-native-ref code pos) -8)
+                                #xfff))
+                 (nlocals (ash (bytevector-u32-native-ref code pos) -20)))
+             (+ nargs nlocals))))
+      ((call call-label tail-call tail-call-label expand-apply-argument)
+       #'(lambda (code pos size) #f))
+      ((shuffle-down)
+       #'(lambda (code pos size)
+           (let ((from (logand (ash (bytevector-u32-native-ref code pos) -8)
+                               #xfff))
+                 (to (ash (bytevector-u32-native-ref code pos) -20)))
+             (and size (- size (- from to))))))
+      (else
+       #f))))
 
 (define (instruction-stack-size-after code pos size)
   (let ((opcode (logand (bytevector-u32-native-ref code pos) #xff)))
-    ((vector-ref stack-effect-parsers opcode) code pos size)))
+    (match (vector-ref stack-effect-parsers opcode)
+      (#f size)
+      (proc (proc code pos size)))))
 
-(define-syntax define-clobber-parser
-  (lambda (x)
-    (syntax-case x ()
-      ((_ name opcode kind arg0 arg* ...)
-       (case (syntax->datum #'kind)
-         ((!)
-          (case (syntax->datum #'name)
-            ((call call-label)
-             #'(let ((parse (lambda (code pos nslots-in nslots-out)
-                              (call-with-values
-                                  (lambda ()
-                                    (disassemble-one code (/ pos 4)))
-                                (lambda (len elt)
-                                  (match elt
-                                    ((_ proc . _)
-                                     (let lp ((slot (- proc 2)))
-                                       (if (< slot nslots-in)
-                                           (cons slot (lp (1+ slot)))
-                                           '())))))))))
-                 (vector-set! clobber-parsers opcode parse)))
-            (else
-             #'(begin))))
-         ((<-)
-          #`(let ((parse (lambda (code pos nslots-in nslots-out)
-                           (call-with-values
-                               (lambda ()
-                                 (disassemble-one code (/ pos 4)))
-                             (lambda (len elt)
-                               (match elt
-                                 ((_ dst . _)
-                                  #,(case (syntax->datum #'arg0)
-                                      ((X8_F24 X8_F12_F12)
-                                       #'(list dst))
-                                      (else
-                                       #'(list (- nslots-out 1 dst)))))))))))
-              (vector-set! clobber-parsers opcode parse)))
-         (else (error "unexpected instruction kind" #'kind)))))))
-
-(define clobber-parsers
-  (make-vector 256 (lambda (code pos nslots-in nslots-out) '())))
-(visit-opcodes define-clobber-parser)
+(define-op-handlers clobber-parsers
+  (lambda (name kind word-types)
+    (match kind
+      ('!
+       (case name
+         ((call call-label)
+          #'(lambda (code pos nslots-in nslots-out)
+              (call-with-values
+                  (lambda ()
+                    (disassemble-one code (/ pos 4)))
+                (lambda (len elt)
+                  (define frame-size 3)
+                  (match elt
+                    ((_ proc . _)
+                     (let lp ((slot (- proc frame-size)))
+                       (if (and nslots-in (< slot nslots-in))
+                           (cons slot (lp (1+ slot)))
+                           '()))))))))
+         (else #f)))
+      ('<-
+       #`(lambda (code pos nslots-in nslots-out)
+           (call-with-values (lambda ()
+                               (disassemble-one code (/ pos 4)))
+             (lambda (len elt)
+               (match elt
+                 ((_ dst . _)
+                  #,(match word-types
+                      (((or 'X8_F24 'X8_F12_F12) . _)
+                       #'(list dst))
+                      (else
+                       #'(if nslots-out
+                             (list (- nslots-out 1 dst))
+                             '()))))))))))))
 
 (define (instruction-slot-clobbers code pos nslots-in nslots-out)
   (let ((opcode (logand (bytevector-u32-native-ref code pos) #xff)))
-    ((vector-ref clobber-parsers opcode) code pos nslots-in nslots-out)))
+    (match (vector-ref clobber-parsers opcode)
+      (#f '())
+      (proc (proc code pos nslots-in nslots-out)))))
